# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/action/save_object.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2
from syft.proto.core.store import (
    store_object_pb2 as proto_dot_core_dot_store_dot_store__object__pb2,
)

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n/proto/core/node/common/action/save_object.proto\x12\x1csyft.core.node.common.action\x1a#proto/core/store/store_object.proto\x1a\x1bproto/core/io/address.proto"h\n\x10SaveObjectAction\x12,\n\x03obj\x18\x01 \x01(\x0b\x32\x1f.syft.core.store.StorableObject\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)


_SAVEOBJECTACTION = DESCRIPTOR.message_types_by_name["SaveObjectAction"]
SaveObjectAction = _reflection.GeneratedProtocolMessageType(
    "SaveObjectAction",
    (_message.Message,),
    {
        "DESCRIPTOR": _SAVEOBJECTACTION,
        "__module__": "proto.core.node.common.action.save_object_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.SaveObjectAction)
    },
)
_sym_db.RegisterMessage(SaveObjectAction)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SAVEOBJECTACTION._serialized_start = 147
    _SAVEOBJECTACTION._serialized_end = 251
# @@protoc_insertion_point(module_scope)
