# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/parameter.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1fproto/lib/torch/parameter.proto\x12\x0esyft.lib.torch"}\n\x0eParameterProto\x12\x0e\n\x06tensor\x18\x01 \x01(\x0c\x12\x15\n\rrequires_grad\x18\x02 \x01(\x08\x12\x11\n\x04grad\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x18\n\x0bgrad_sample\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x42\x07\n\x05_gradB\x0e\n\x0c_grad_sampleb\x06proto3'
)


_PARAMETERPROTO = DESCRIPTOR.message_types_by_name["ParameterProto"]
ParameterProto = _reflection.GeneratedProtocolMessageType(
    "ParameterProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _PARAMETERPROTO,
        "__module__": "proto.lib.torch.parameter_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.ParameterProto)
    },
)
_sym_db.RegisterMessage(ParameterProto)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PARAMETERPROTO._serialized_start = 51
    _PARAMETERPROTO._serialized_end = 176
# @@protoc_insertion_point(module_scope)
