# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/tensor/fixed_precision_tensor.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.tensor import (
    share_tensor_pb2 as proto_dot_core_dot_tensor_dot_share__tensor__pb2,
)
from syft.proto.core.tensor import (
    tensor_pb2 as proto_dot_core_dot_tensor_dot_tensor__pb2,
)
from syft.proto.lib.numpy import array_pb2 as proto_dot_lib_dot_numpy_dot_array__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n.proto/core/tensor/fixed_precision_tensor.proto\x12\x10syft.core.tensor\x1a\x1eproto/core/tensor/tensor.proto\x1a$proto/core/tensor/share_tensor.proto\x1a\x1bproto/lib/numpy/array.proto"\xc8\x01\n\x14\x46ixedPrecisionTensor\x12*\n\x06tensor\x18\x01 \x01(\x0b\x32\x18.syft.core.tensor.TensorH\x00\x12.\n\x05share\x18\x02 \x01(\x0b\x32\x1d.syft.core.tensor.ShareTensorH\x00\x12+\n\x05\x61rray\x18\x03 \x01(\x0b\x32\x1a.syft.lib.numpy.NumpyProtoH\x00\x12\x0c\n\x04\x62\x61se\x18\x04 \x01(\r\x12\x11\n\tprecision\x18\x05 \x01(\rB\x06\n\x04\x64\x61tab\x06proto3'
)


_FIXEDPRECISIONTENSOR = DESCRIPTOR.message_types_by_name["FixedPrecisionTensor"]
FixedPrecisionTensor = _reflection.GeneratedProtocolMessageType(
    "FixedPrecisionTensor",
    (_message.Message,),
    {
        "DESCRIPTOR": _FIXEDPRECISIONTENSOR,
        "__module__": "proto.core.tensor.fixed_precision_tensor_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.tensor.FixedPrecisionTensor)
    },
)
_sym_db.RegisterMessage(FixedPrecisionTensor)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _FIXEDPRECISIONTENSOR._serialized_start = 168
    _FIXEDPRECISIONTENSOR._serialized_end = 368
# @@protoc_insertion_point(module_scope)
