# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/adp/scalar_manager.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n#proto/core/adp/scalar_manager.proto\x12\rsyft.core.adp\x1a\x1bproto/lib/python/dict.proto"\x0e\n\x0cPrimeFactory"\x85\x01\n"VirtualMachinePrivateScalarManager\x12\x32\n\rprime_factory\x18\x01 \x01(\x0b\x32\x1b.syft.core.adp.PrimeFactory\x12+\n\x0cprime2symbol\x18\x02 \x01(\x0b\x32\x15.syft.lib.python.Dictb\x06proto3'
)


_PRIMEFACTORY = DESCRIPTOR.message_types_by_name["PrimeFactory"]
_VIRTUALMACHINEPRIVATESCALARMANAGER = DESCRIPTOR.message_types_by_name[
    "VirtualMachinePrivateScalarManager"
]
PrimeFactory = _reflection.GeneratedProtocolMessageType(
    "PrimeFactory",
    (_message.Message,),
    {
        "DESCRIPTOR": _PRIMEFACTORY,
        "__module__": "proto.core.adp.scalar_manager_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.PrimeFactory)
    },
)
_sym_db.RegisterMessage(PrimeFactory)

VirtualMachinePrivateScalarManager = _reflection.GeneratedProtocolMessageType(
    "VirtualMachinePrivateScalarManager",
    (_message.Message,),
    {
        "DESCRIPTOR": _VIRTUALMACHINEPRIVATESCALARMANAGER,
        "__module__": "proto.core.adp.scalar_manager_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.adp.VirtualMachinePrivateScalarManager)
    },
)
_sym_db.RegisterMessage(VirtualMachinePrivateScalarManager)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _PRIMEFACTORY._serialized_start = 83
    _PRIMEFACTORY._serialized_end = 97
    _VIRTUALMACHINEPRIVATESCALARMANAGER._serialized_start = 100
    _VIRTUALMACHINEPRIVATESCALARMANAGER._serialized_end = 233
# @@protoc_insertion_point(module_scope)
