# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/io/address.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.io import location_pb2 as proto_dot_core_dot_io_dot_location__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1bproto/core/io/address.proto\x12\x0csyft.core.io\x1a\x1cproto/core/io/location.proto"\xb3\x02\n\x07\x41\x64\x64ress\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x13\n\x0bhas_network\x18\x03 \x01(\x08\x12/\n\x07network\x18\x04 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x12\n\nhas_domain\x18\x05 \x01(\x08\x12.\n\x06\x64omain\x18\x06 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x12\n\nhas_device\x18\x07 \x01(\x08\x12.\n\x06\x64\x65vice\x18\x08 \x01(\x0b\x32\x1e.syft.core.io.SpecificLocation\x12\x0e\n\x06has_vm\x18\t \x01(\x08\x12*\n\x02vm\x18\n \x01(\x0b\x32\x1e.syft.core.io.SpecificLocationb\x06proto3'
)


_ADDRESS = DESCRIPTOR.message_types_by_name["Address"]
Address = _reflection.GeneratedProtocolMessageType(
    "Address",
    (_message.Message,),
    {
        "DESCRIPTOR": _ADDRESS,
        "__module__": "proto.core.io.address_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.io.Address)
    },
)
_sym_db.RegisterMessage(Address)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _ADDRESS._serialized_start = 76
    _ADDRESS._serialized_end = 383
# @@protoc_insertion_point(module_scope)
