# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/python/bytes.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1cproto/lib/python/bytes.proto\x12\x0fsyft.lib.python"\x15\n\x05\x42ytes\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x62\x06proto3'
)


_BYTES = DESCRIPTOR.message_types_by_name["Bytes"]
Bytes = _reflection.GeneratedProtocolMessageType(
    "Bytes",
    (_message.Message,),
    {
        "DESCRIPTOR": _BYTES,
        "__module__": "proto.lib.python.bytes_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.python.Bytes)
    },
)
_sym_db.RegisterMessage(Bytes)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _BYTES._serialized_start = 49
    _BYTES._serialized_end = 70
# @@protoc_insertion_point(module_scope)
